; this is a translation of the pascal squeeze program into ASM

; CP/M compatible file squeezer utility.
;
; This translation uses the Huffman algorithm to develop a binary tree representing the decoding information
; for a variable length bit string code for each input value.  Each string's length is in inverse proportion
; to its frequency of appearance in the incoming data stream.  The encoding table is derived from the decoding
; table.
;
; The range of valid values into the Huffman algorithm are the values of a byte stored in an integer plus the
; special endfile value chosen to be an adjacent value. Overall, 0-SPEOF.
;
; The algorithm develops the single element trees into a single binary tree by forming subtrees rooted in
; interior nodes having weights equal to the sum of weights of all their descendents and having depth counts
; indicating the depth of their longest paths.
;
; When all trees have been formed into a single tree satisfying the heap property (on weight, with depth as a
; tie breaker) then the binary code assigned to a leaf (value to be encoded) is then the series of left (0) and
; right (1) paths leading from the root to the leaf.  Note that trees are removed from the heaped list by moving
; the last element over the top element and reheaping the shorter list.
;
; To further compress the output code stream, all bytes pass directly through except for:
;       1) DLE is encoded as (DLE, zero).
;       2) repeated byte values (count >= 3) are encoded as (value, DLE, count).
;
; In the original design it was believed that a Huffman code would fit in the same number of bits that will hold
; the sum of all the counts.  That was disproven by a user's file and was a rare but infamous bug. This version
; attempts to choose among equally weighted subtrees according to their maximum depths to avoid unnecessarily long
; codes. In case that is not sufficient to guarantee codes <= 16 bits long, we initially scale the counts so the
; total fits in an unsigned integer, but if codes longer than 16 bits are generated the counts are rescaled to a
; lower ceiling and code generation is retried.
;
; The "node" array of structures contains the nodes of the binary tree. The first NUMVALS nodes are the leaves
; of the tree and represent the values of the data bytes being encoded and the special endfile, SPEOF.  The
; remaining nodes become the internal nodes of the tree.
;
; Program states:
;    NoHist    don't consider previous input
;    SentChar  lastchar set, no lookahead yet
;    SendNewC  newchar set, previous sequence done
;    SendCnt   newchar set, DLE sent, send count next
;
print macro string
   lea   dx,string
   mov   ah,9
   int   21h
   endm
cseg segment para public 'CODE'
   assume  cs:cseg,ds:cseg
   org   100h
entry:
     jmp    main
hello db "File squeeze program",13,10,'$'
istack equ this word
buffer1   dw 0
buffer2   dw 0
infile db 64 dup(0)
outfile db 64 dup(0)
true  equ 0
false equ -1
space     equ  ' '
Error     equ -1
Null      equ -2
Recognize equ 0FF76h;  { unlikely pattern }
DLE       equ 90h;
SPEOF     equ 256;    { special endfile token }
NumVals   equ 257;    { 256 data values plus SPEOF }
NoChild   equ -1;     { indicates end of path through tree }
maxcount  equ 07fffh ; { biggest UNSIGNED integer }

;        ValType  = array[0..numvals] of integer;
;tateTypes dw NoHist,SentChar,SendNewC,SendCnt,EndFile
     nohist  equ  0
     sentchar equ 1
     sendnewc equ 2
     sendcnt  equ 3
     endfile  equ 4
;
        NodeType struc
             weight dw 0  ;             { number of appearances }
             tdepth dw 0  ;          { length on longest path in tree }
             lchild dw 0  ;
             rchild dw 0  ;  { indices to next level }
        NodeType ends

        InFileName  db 64 dup(0)
        OutFileName db 64 dup(0)
        start       dw  0
        finish      dw  0
        i           dw  0
        crc         dw  0    ;  { Cyclic Redundancy Check code }
        likect      dw  0    ;   { count of consecutive identical chars }
        lastchar    db  0
        newchar     db  0

        EOFlag      db  0
        done        db  0

NumNodesc equ 513;    { equ 2 * NUMVALS - 1 equ number of nodes }
        numnodes   dw numnodesc
        node       dw numnodesc dup(0,0,0,0)
        dctreehd    dw  0 ;   { index to head node of final tree }
        state   dw  0
        l       dw  0
        top     dw  0
        bottom  dw  0
        r       dw  0
        list    dw  1024 dup(0)

; This is the encoding table:  The bit strings have first bit in = low bit.
; Note that counts were scaled so code fits UNSIGNED integer }
        codelen   dw numvals dup(0)
        codex     dw numvals dup(0)
;       codelen, code: array[0..numvals] of integer; { number of bits in code & code itself, right adjusted }
        tcode dw  0  ;    { temporary code value }

        curin dw  0  ;    { Value currently being encoded }
        cbitsrem dw  0  ; { Number of code string bits remaining }
        ccode dw  0  ;    { Current code shifted so next code bit is at right }


zero_tree proc near
;   { Initialize all nodes to single element binary trees with zero weight and depth. }
    push   bx
    push  cx
    mov  cx,numnodes
    mov  bx,0
zt_loop:
    mov   node[bx].weight,0;
    mov   node[bx].tdepth,0;
    mov   node[bx].lchild,NoChild;
    mov   node[bx].rchild,NoChild;
    add   bx,8
    loop  zt_loop
    pop  cx
    pop  bx
     ret
zero_tree endp
;


putwe  proc     near
; character passed in ax
; { write out low order byte of word to file, then high order byte regardless of host CPU. }
    stosw
    cmp   di,4000h
    jl    ok_store
    call  flush_buffer
    mov   di,0
ok_store:
    ret
;   var b1, b2: char;
;   begin
;     b1 := chr(w and $FF);
;     b2 := chr(w shr 8);
;     write(OutFile,b1,b2);
;   end;
putwe  endp


GetC_CRC proc near
;unction GetC_CRC: char;
;  { Get next byte from file and update checksum }
   cmp   si,bp   ; bp has the current size of file
   jg    ok_buffer
   call  fill_buffer
   jnc   ok_buffer
   ret
ok_buffer:
   lodsb
   mov   ah,0
   add   crc,ax
   ret  ; with character in al
GetC_CRC endp
;

;   var c: char;
;   begin
;     if not(eof(InFile))
;       then begin
;              read(InFile,c);
;              crc := crc + ord(c); { update checksum }
;            end
;       else EOFlag := true;
;     GetC_CRC := c;  {undefined if EOFlag is true}
;   end;


PrintBits  proc near ;PrintBits(len, number: integer);
; I think this is debug code
;   var i, j: integer;
;   begin
;     write('  code ');
;     for i:=len-1 downto 0
;       do begin
;            j := (number shr i) and $0001;
;            write(j:1);
;          end;
;     writeln;
;   end; *)
     ret
Printbits endp

;
alike:  ; set carry, carry = flase
        call       getc_crc;
        jnc        ok_get1
alikef:
        stc
        ret
ok_get1:
        cmp     lastchar,al
        jne   alikef
        cmp    likect,255
        jge    alikef
        clc
        ret
;         else begin
;                if (newchar = lastchar) and (likect < 255)
;                  then alike := true
;                  else alike := false;
;              end;
;     end;

NoHistory    proc near ;NoHistory; {set up the state machine}
     mov   state,SentChar;
     call  GetC_CRC
     mov lastchar,al
     jc   NoHist_e
     ret
NoHist_e:
     mov state,EndFile;
     clc
     ret
NoHistory    endp

SentAChar    proc near  ;SentAChar;   {Lastchar is set, need lookahead}

SentDLE:
;       begin
    mov   state,NoHist;
    mov   ax,0
    ret
SentAChar    endp

CheckAlike  proc near ;CheckAlike;
      mov    likect,1;
;  while alike do likect := likect + 1;
while_alike:
      call   alike
      jc     not_alike
      inc    likect
      cmp    likect,1
      je     alike1
      cmp    likect,2
      je     alike1
      mov   state,SendCnt;
      mov   ax,DLE
;         end;
      jmp    while_alike
alike1:
;           1: begin
;                lastchar := newchar;
;                return := lastchar;
;              end;
           mov   lastchar,al
      jmp    while_alike
alike2:
;           2: begin { just pass through }
             mov state,SendNewC;
             mov   al,lastchar
;                return := lastchar;
;              end;
;       end;
      jmp    while_alike
not_alike:

;     begin
        cmp EOFlag,1
        je   alike_end

;         else begin
        cmp    lastchar,DLE
        je   SentDLE
        jmp  CheckAlike;
;              end;
;     end;
alike_end:
;       if EOFlag
;         then state := EndFile {no return value, set to SPEOF in calling routine}
       mov state,endfile
       stc
       ret
CheckAlike  endp

SendNewChar  proc near;   {Previous sequence complete, newchar set}
    mov state,SentChar;
    mov lastchar,al
   ;    return := lastchar;
    ret  ; al has the character
SendNewChar  endp

SendCount  proc near;  {Sent DLE for repeat sequence, send count}
   mov  state,SendNewC;
   call  likect
      ; return := chr(likect);
   ret
SendCount  endp

getcnr  proc  near  ; returns something in al
; function getcnr: char;
;   var return: char;
    push   bx
    mov    bx,state
    shl    bx,1
    call   statelist[bx]
    pop    bx
;     case state of
;       NoHist:   NoHistory;
;       SentChar: SentAChar;
;       SendNewC: SendNewChar;
;       SendCnt:  SendCount;
;     else writeln('program bug - bad state');
;     end;
;     getcnr := return;
;   end;
     ret   ; something in al
getcnr  endp        ; returns something in al

statelist  equ this word
        dw        NoHistory;
        dw        SentAChar;
        dw        SendNewChar;
        dw        SendCount;
;       dw        error_out


Write_header  proc near  ;
;  { Write out the header of the compressed file }
;
;    var i, k, l, r, numnodes: integer;
;        { numnodes: nbr of nodes in simplified tree }
;
;   begin
; pass character to putwe in al
;     putwe(RECOGNIZE);   { identifies as compressed }
      mov   ax,recognize
      call   putwe
;     putwe(crc);         { unsigned sum of original data }
      mov   ax,crc
      call   putwe

;     { Record the original file name w/o drive }
;     if (InFileName[2] = ':')
;       then InFileName := copy(InFileName,3,length(InFileName)-2);
      push  si
      push  di
      lea   si,InfileName
      mov   di,si
      cmp   Infilename+1,':'
      jne   no_colon
      inc   si
      inc   si
      mov   di,si
no_colon:
      cld
      mov    cx,62
      mov  al,'\'
check_slash:
repne  scasb
      jne  no_bslash
      mov  si,di
      jmp  check_slash
no_bslash:
      mov  si,di
      pop  di
      push di
      mov  cx,12
      cld
rep  movsb
     pop  di
     mov  al,0
repne scasb
     pop  si
;
;     InFileName := InFileName + chr(0);  {mark end of file name}
;     for i:=1 to length(InFileName) do write(OutFile,InFileName[i]);
;
;     { Write out a simplified decoding tree. Only the interior nodes are written. When a child is a leaf index
;       (representing a data value) it is recoded as -(index + 1) to distinguish it from interior indexes which
;       are recoded as positive indexes in the new tree.  Note that this tree will be empty for an empty file. }

;     if dctreehd < NUMVALS
      cmp  dctreehd,NUMVALS
      jl   low_num
;       else numnodes := dctreehd - (NUMVALS - 1);
      mov  ax,dctreehd
      sub  ax,NUMVALS-1
      jmp  by_val
low_num:
;       then numnodes := 0
       mov numnodes,0
by_val:
;   pass character to putwe in al
;     putwe(numnodes);
      mov  ax,numnodes
      call  putwe

;     i := dctreehd;  use bx for index
      mov  ax,dctreehd
      mov bx,ax
      shl bx,1
      shl bx,1
      shl bx,1
;     for k:=0 to numnodes-1
      mov  cx,numnodes
      dec  cx
do_begin:
      mov  ax,node[bx].lchild;
      mov  l,ax
      mov  ax,node[bx].rchild;
      mov  r,ax
;            if l < NUMVALS
      cmp  l,numvals
      jl   lo_l
;              else l := dctreehd - l;
      mov   ax,dctreehd
      sub   ax,l
      mov   l,ax
      jmp  by_l
lo_l:
;        then l := -(l + 1)
      inc  l
      neg  l
by_l:
;            if r < NUMVALS
      cmp  l,numvals
      jl   low_r
      mov   ax,dctreehd
      sub   ax,r
      mov   r,ax
      jmp  by_r
;              else r := dctreehd - r;
low_r:
;              then r := -(r + 1)
      inc  r
      neg  r
by_r:
; pass character to putwe in al
;            putwe(l);  { left child }
      mov  ax,l
      call  putwe
;            putwe(r);  { right child }
      mov  ax,r
      call  putwe
      dec   bx
      loop  do_begin
;            i := i - 1;
;          end;
;   end;
       ret

Write_header  endp

;
cmptrees: ;(a, b: integer): boolean; {entry with root nodes}
;    { Compare two trees, if a > b return true, else return false. }
; enters si and di with things to compare
;     begin
;       cmptrees := false;
;       if node[a].weight > node[bx].weight
        mov ax,node[si].weight
        cmp   ax,node[di].weight
        jle   ct_true
;
;         else if node[a].weight = node[b].weight
        jne ct_false
;                then if node[a].tdepth > node[b].tdepth
          mov  ax,node[si].tdepth
          cmp  ax,node[di].tdepth
          jle  ct_true
ct_false:
    stc
    ret
ct_true:
;         then cmptrees := true
    clc
    ret
Adjust:  ; (top, bottom dw  0  ;var list: ValType);
;  { Make a heap from a heap with a new top }
;
;    var k, temp: integer;

;   begin
;     k := 2 * top + 1;    { left child of top }
    mov  bx,top
    shl  bx,1
    inc  bx
;     temp := list[top];   { remember root node of top tree }
    mov  dx,list[bx]
    cmp  dx,bottom
    jle  adj1
    ret
adj1:
   push  bx
;  if (k <= bottom)
;  then begin
;    if (k < bottom) and (cmptrees(list[k], list[k+1])) then k :=k+1;
    mov   bp,list[bx]
    cmp  dx,bottom
    jge   adj2
adj2:
;      { k indexes "smaller" child (in heap of trees) of top
;        now make top index "smaller" of old top and smallest child }
;        if cmptrees(temp,list[k])
;        then begin
;               list[top] := list[k];
;               list[k] := temp;
;               adjust(k, bottom, list);
;                     end;
;            end;
;   end;

main:
  print mello
     call parse
     lea  sp,istack
     call shrink
     int  20h


  { get filename to process & convert to upper case}
  write('Enter file to squeeze: ');   readln(InFileName);   writeln;
  for i:=1 to length(InFileName) do InFileName[bx] := upcase(InFileName[bx]);

  { Find and change output file type }
  start := 1; { skip leading blanks }
  while (InFileName[start]=space) and (start <= length(InFileName)) do start := start + 1;
  InFileName := copy(InFileName, start, length(InFileName)-start+1);
  finish := pos('.',InFileName);
  if finish=0
    then OutFileName := InFileName + '.QQQ'
    else begin
           OutFileName := InFileName;
           OutFileName[finish+2] := 'Q';
         end;

  { open source file and check for existence }
  assign(InFile,InFileName);   assign(OutFile,OutFileName);
  {$I-}   reset(InFile);   {$I+}
  if IOresult=0
    then begin
           write('The file ',InFileName,' (',longfilesize(InFile):6:0);
           writeln(' bytes) is being squeezed to ',OutFilename);
           Initialize_Huffman;
           squeeze;
           writeln(', Done.');   close(InFile);   close(OutFile);
         end
    else writeln('Error -- input file doesn''t exist');

end.

flush_buffer:
fill_buffer:
last_byte db 16 dup(0)
error2  db 'Error allocating memory',13,10,'$'
shrink:
public shrink
; allocates memory
; shrink
;
     mov  ah,4ah
     lea  bx,last_byte
     shr  bx,1
     shr  bx,1
     shr  bx,1
     shr  bx,1
     inc  bx   ; for safety
     int  21h
     jnc  good_shrink
bad_mem:
     print error2
     stc
     ret
good_shrink:
; buffer1=32k
     mov  ah,48h
     mov  bx,8000h ; 20k
     int  21h
     jc   bad_mem
     mov  buffer1,ax
; buffer2=32k
     mov  ah,48h
     mov  bx,8000h ; 20k
     int  21h
     jc   bad_mem
     mov  buffer2,ax
     clc
     ret
cseg  ends
      end  entry

  function buildenc(level, root: integer) dw  0  ;{returns error or null}

  { Recursive routine to walk the indicated subtree and level
    and maintain the current path code in bstree. When a leaf
    is found the entire code string and length are put into
    the encoding table entry for the leaf's data value.

    Returns ERROR if codes are too long. }

    var l, r, return: integer;

    begin
      return := null;
      l := node[root].lchild;
      r := node[root].rchild;

      if (l=NOCHILD) and (r=NOCHILD)
        then begin  {have a leaf}
               codelen[root] := level;
               codex[root] := tcode and ($FFFF shr (16 - level));
               if level > 16
                 then return := ERROR
                 else return := NULL;
             end
        else begin
               if l <> NOCHILD
                 then begin  {Clear path bit and go deeper}
                        tcode := tcode and not(1 shl level);
                        if buildenc(level+1,l) = ERROR then return := ERROR;
                      end;
               if r <> NOCHILD
                 then begin  {Set path bit and go deeper}
                        tcode := tcode or (1 shl level);
                        if buildenc(level+1,r)=ERROR then return := ERROR;
                      end;
             end;
      buildenc := return;
    end;
{.pa}
  proc Build_Tree(var list: ValType; len: integer); {Huffman algorithm}

    var freenode dw  0  ;        {next free node in tree}
        lch, rch dw  0  ;        {temporaries for left, right children}
        i: integer;

    function Maximum(a, b: integer): integer;
      begin
        if a>b then Maximum:=a else Maximum:=b;
      end;

    begin
      write(', Building tree');
      { Initialize index to next available (non-leaf) node.
        Lower numbered nodes correspond to leaves (data values). }
      freenode := NUMVALS;

      { Take from list two btrees with least weight and build an
        interior node pointing to them.  This forms a new tree. }
      while (len > 1)
        do begin
             lch := list[0]; { This one will be left child }

             { delete top (least) tree from the list of trees }
             len := len - 1;
             list[0] := list[len];
             adjust(0, len - 1, list);

             { Take new top (least) tree. Reuse list slot later }
             rch := list[0]; { This one will be right child }

             { Form new tree from the two least trees using a free node as root.
               Put the new tree in the list. }
             with node[freenode]
              do begin;
                   lchild := lch;
                   rchild := rch;
                   weight := node[lch].weight + node[rch].weight;
                   tdepth := 1 + Maximum(node[lch].tdepth, node[rch].tdepth);
                 end;
             list[0] := freenode;       {put at top for now}
             freenode := freenode + 1;  {next free node}
             { reheap list to get least tree at top }
             adjust(0, len - 1, list);
           end;
      dctreehd := list[0];   { head of final tree }
    end;
{.pa}
  proc Initialize_Huffman;

  { Initialize the Huffman translation. This requires reading the input file through any preceding translation
    functions to get the frequency distribution of the various values. }

    var c, i: integer;
        btlist: ValType;   { list of intermediate binary trees }
        listlen dw  0  ; { length of btlist }
        ceiling dw  0  ; { limit for scaling }

{ Heap and Adjust maintain a list of binary trees as a heap with the top indexing the binary tree on the list which
  has the least weight or, in case of equal weights, least depth in its longest path. The depth part is not strictly
  necessary, but tends to avoid long codes which might provoke rescaling. }

    proc Heap(var list: ValType; l: integer);
      var i, len: integer;
      begin
        len := (l - 2) div 2;
        for i:=len downto 0 do adjust(i, l - 1, list);
      end;
(*
    proc PrintFrequency;
      var i, j: integer;
      begin
        j := 0;
        for i:=0 to numvals-1
          do if node[bx].weight>0
               then begin
                      j := j + 1;
                      writeln(lst,'node ',i:3,'  weight is ',node[bx].weight:4:0);
                    end;
        writeln(lst);
        writeln(lst,'Total node count is ',j);
      end;

    proc PrintList;
      var i: integer;
          str: string[10];
      begin
        writeln(', waiting');   readln(str);
        for i:=0 to numvals-1
          do begin
               write('number ',i:3,'  length ',codelen[bx]:2);
               write('  weight ',node[bx].weight:4:0);
               if codelen[bx]>0 then PrintBits(codelen[bx], codex[bx]) else writeln;
             end;
      end;
  *)
  begin
      write('Pass 1: Analysis');
      crc := 0;   zero_tree;   state := NoHist;   EOFlag := false;

      repeat    { Build frequency info in tree }
          c := ord(getcnr);
          if EOFlag then c := SPEOF;
          with node[c] do if weight < maxcount then weight := weight + 1;
          if EOFlag then write(', End of file found');
        until (EOFlag);
      {PrintFrequency;}

      ceiling := MAXCOUNT;

      { Try to build encoding table. Fail if any code is > 16 bits long. }
      repeat
          if (ceiling <> MAXCOUNT) then write('*** rescaling ***, ');
          scale(ceiling);
          ceiling := ceiling div 2;  {in case we rescale again}

          listlen := 0;   {find length of list and build single nodes}
          for i:=0 to numvals-1
            do begin
                 if node[bx].weight > 0
                   then begin
                          node[bx].tdepth := 0;
                          btlist[listlen] := i;
                          listlen := listlen + 1;
                        end;
               end;
          heap(btlist, listlen-1);  { *** changed from listlen }
          Build_Tree(btlist, listlen);
          for i := 0 to NUMVALS-1 do codelen[bx] := 0;
        until (buildenc(0,dctreehd) <> ERROR);

      {PrintList;}
      { Initialize encoding variables }
      cbitsrem := 0;   curin := 0;
    end;
{.pa}
  function gethuff: char; {returns byte values except for EOF}
  { Get an encoded byte or EOF. Reads from specified stream AS NEEDED.

    There are two unsynchronized bit-byte relationships here:
      The input stream bytes are converted to bit strings of various lengths via
      the static variables named Cxxxxx.  These bit strings are concatenated without
      padding to become the stream of encoded result bytes, which this function
      returns one at a time. The EOF (end of file) is converted to SPEOF for
      convenience and encoded like any other input value. True EOF is returned after
      that. }

    var rbyte dw  0  ;      {Result byte value}
        need, take dw  0  ; {numbers of bits}
        return: integer;

    begin
      rbyte := 0;
      need := 8;        {build one byte per call}
      return := ERROR;  {start off with an error}

      {Loop to build a byte of encoded data.  Initialization forces read the first time}
      while return=ERROR
        do begin
             if cbitsrem >= need
               then begin {Current code fullfills our needs}
                      if need = 0
                        then return := rbyte and $00FF
                        else begin
                               rbyte := rbyte or (ccode shl (8 - need)); {take what we need}
                               ccode := ccode shr need;                  {and leave the rest}
                               cbitsrem := cbitsrem - need;
                               return := rbyte and $00FF;
                             end;
                    end
               else begin
                      if cbitsrem > 0
                        then begin  {We need more than current code}
                               rbyte := rbyte or (ccode shl (8 - need)); {take what there is}
                               need := need - cbitsrem;
                             end;
                      if curin = SPEOF
                        then begin
                               cbitsrem := 0;
                               if need=8
                                 then begin                       {end of file}
                                        done := true;
                                        return := 0; {any valid char value}
                                      end
                                 else return := rbyte and $00FF;  {data first}
                             end
                        else begin
                               curin := ord(getcnr);
                               if EOFlag then curin := SPEOF;
                               ccode := codex[curin];
                               cbitsrem := codelen[curin];
                             end;
                    end;
           end;
      gethuff := chr(return);
    end;
{.pa}
  proc squeeze;
    var c: char;
    begin
      writeln;   write('Pass 2: Squeezing');
      reset(InFile);   rewrite(OutFile);   EOFlag := false;
      write(', header');   Write_Header;
      write(', body');     state := NoHist;
      done := false;   c := gethuff;  {prime while loop}
      while not(done)
        do begin
             write(OutFile,c);
             c := gethuff;
           end;
    end;

; scaling is no needed because we are reading in 5k chunks at a time
; The count of number of occurrances of each input value have already been prevented from exceeding MAXCOUNT.
; Now we must scale them so that their sum doesn't exceed ceiling and yet no non-zero count can become zero.
; This scaling prevents errors in the weights of the interior nodes of the Huffman tree and also ensures that
; the codes will fit in an unsigned integer.  Rescaling is used if necessary to limit the code length. }
;
Scale; (ceil: integer);  { upper limit on total weight }

    var i, c, ovflw, divisor: integer;
        w, sum: real;
        increased: boolean;

    begin
      repeat
          sum := 0;   ovflw := 0;
          for i:=0 to numvals-1
            do begin
                 if node[bx].weight > (ceil - sum) then ovflw := ovflw + 1;
                 sum := sum + node[bx].weight;
               end;

          divisor := ovflw + 1;

          { Ensure no non-zero values are lost }
          increased := FALSE;
          for i:=0 to numvals-1
            do begin
                 w := node[bx].weight;
                 if (w < divisor) and (w <> 0)
                   then begin
                          { Don't fail to provide a code if it's used at all }
                          node[bx].weight := divisor;
                          increased := TRUE;
                        end;
               end;
        until not(increased);

      { Scaling factor choosen, now scale }
      if divisor > 1
        then for i:=0 to numvals-1
               do with node[bx] do weight := int((weight / divisor) + 0.5);
    end;
{.pa}
